;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; CHECKSUM.S - macros & routines for generating and checking checksums

; each type is in two parts, 
;	A routine to generate the checksum given a0,a1,a2 as start,end,dest
; 	A macro to check the checksum and crash if wrong
; 	the macros will embed information to enable the mastering
;	program to calculate the checksum
;	the markers must be well disguised

	ifnd	base
base	equ	$1000
	endc

c_marker	equ	"CARK"
r_marker	equ	"RARK"

	ifd	lethal_weapon
parasol_stars	equ	1
	endc

	include	c:\dev\pushpop.s

	ifd	parasol_stars
	pushs
	section		cs,text cs_group
checksum_list
	dc.l	c_marker
	section		rt,text	rt_group
runtime_list
	dc.l	r_marker
	pops
	endc


marker	macro	type,start,end,dest
;	dc.l	c_marker
	ifd	parasol_stars
	pushs
	section	cs,text	cs_group
	dc.w	\1	     			; checksum type
	dc.l	\2,\3,\4	 		; start,end,crash
	pops
	endc
	endm

rmarker	macro	type,start,end,dest
;	dc.l	r_marker
	pushs
	section	rt,text rt_group
	dc.w	\1
	dc.l	\2,\3,\4
	pops
	endm


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 1 - A fairly simple one, relative position dependent

	ifd	mastering
do_checksum_1
	clr.l	d0
.next
	add.b	(A0)+,d0
	ror.l	#1,d0
	cmp.l	a1,a0
	blt	.next
	move.w	d0,(A2)
	rts
	endc

checksum_1	macro
	marker	1,\1,\2,.sum\@
	move.l	#(\1)/2,a0    			; BRACKETS!!!
	lea	\2,a1
	clr.l	d0
	add.l	a0,a0
.next\@
	add.b	(A0)+,d0
	ror.l	#1,d0
	cmp.l	a1,a0
	blt	.next\@
.sum\@	equ	*+2
	sub.w	#5678,d0

	ifeq	checksum_crash
	move.l	#\4,d1
	jsr	maybe_error
	endc

	move.l	#(\3)/2,a0
	add.l	a0,a0
	add.w	d0,(A0)

.ok\@
	endm


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 2 - Does one long word at a time, so can be put in an int routine 
;     (eg keyboard, vbl, or any routine that is only called occasionally

	ifd	mastering
do_checksum_2
	clr.l	d0
	clr.w	d1
.next
	add.b	(A0)+,d0
	ror.w	#1,d0

	cmp.l	a1,a0
	blt	.next
	move.w	d0,(A2)
	rts
	endc

; this macro should go inline in a frequently called routine
; eg vbl or screen flip, or any routine called nearly every game cycle	
checksum_2	macro
	marker	2,\1,\2,.sum\@
	lea	.value\@(pc),a4
	movem.l	(a4),d0/d1/a0-a2
	add.b	(A1)+,d0
	ror.w	#1,d0

	cmp.l	a2,a1
	blt.s	.ok\@
	move.l	a0,a1	
.sum\@	equ	*+2
	sub.w	#1678,d0

	ifeq	checksum_crash
	move.l	#\4,d1
	jsr	maybe_error
	endc

	move.l	#(\3)-24300,a3
	eor.w	d0,24300(A3)		; third parameter must be a vital intruction
	clr.l	d1
	bra	.ok\@		; possible a register list, as difficult to trace

.value\@	dc.l	0
.count\@	dc.l	0
.base\@		dc.l	\1
.now\@		dc.l	\1
.end\@		dc.l	\2

.ok\@
	movem.l	d0/d1/a0-a2,(a4)
	endm


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 3 - Another fairly simple one, (sub.w ror.l) a bit of crc ing

	ifd	mastering
do_checksum_3
	moveq	#3,d0
	move.l	a1,d7
	sub.l	a0,d7
	lsr.w	#1,d7
	subq.w	#1,d7
.next
	sub.w	(A0)+,d0
	ror.l	#1,d0
	dbf	d7,.next
	move.w	d0,(A2)
	rts
	endc

checksum_3	macro
	marker	3,\1,\2,.sum\@
	move.l	#(\1)-2,a0    		
	moveq	#3,d0
	addq.l	#2,a0
	move.w	#((\2)-(\1))/2-1,d7
.next\@
	sub.w	(A0)+,d0
	ror.l	#1,d0
	dbf	d7,.next\@
	move.l	#(\3)-8,a0
.sum\@	equ	*+2
	sub.w	#5678,d0

	ifeq	checksum_crash
	move.l	#\4,d1
	jsr	maybe_error
	endc

	addq.l	#8,a0
	sub.w	d0,(A0)
.ok\@
	endm


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 4 - Sort of CRC checksum (add.b, ror.l) - quite quick

	ifd	mastering
do_checksum_4
	clr.l	d0
.next
	add.b	(A0)+,d0
	ror.l	#1,d0
	cmp.l	a1,a0
	blt.s	.next
	move.w	d0,(A2)
	rts
	endc

checksum_4	macro
	marker	4,\1,\2,.sum\@
	move.l	#(\1)/2,a0    			; BRACKETS!!!
	lea	\2,a1
	clr.l	d0
	add.l	a0,a0
.next\@
	add.b	(A0)+,d0
	ror.l	#1,d0
	cmp.l	a1,a0
	blt.s	.next\@
.sum\@	equ	*+2
	sub.w	#5678,d0

	ifeq	checksum_crash
	move.l	#\4,d1
	jsr	maybe_error
	endc

	move.l	#(\3)-100,a0
	add.w	d0,100(A0)

.ok\@
	endm


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 5 - CRC checksum     (rol.l, sub.b)
; with provision for adding a line in the main loop
; uses d6,d7,a6
; so use this in place of a simple loop, eg to fill a table
; or to copy a string of fixed length, etc
; use:	checksum_5 start,start+len,crash,n,<move.l (a0)+,(a1)+>

	ifd	mastering
do_checksum_5
	moveq	#0,d0
	move.l	a1,d7
	sub.l	a0,d7
	subq.w	#1,d7
.next
	rol.l	#1,d0
	sub.b	(A0)+,d0
	dbf	d7,.next
	move.w	d0,(A2)
	rts
	endc

checksum_5	macro
	marker	5,\1,\2,.sum\@
	lea	\1,a6  
	move.w	#((\2)-(\1))-1,d7  		   
	clr.l	d6
.next\@
	rol.l	#1,d6
	sub.b	(A6)+,d6
	\5
	dbf	d7,.next\@
.sum\@	equ	*+2
	sub.w	#5678,d6

	ifeq	checksum_crash
	pushem
	move.w	d6,d0
	move.l	#\4,d1
	jsr	maybe_error
	popem
	endc

	move.l	#(\3)+2,a6
	add.w	d6,-(A6)

.ok\@
	endm



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 6 - OBBLE checksum  
; must be placed inline in a bit of obble code 
;

	ifd	mastering
do_checksum_6
	moveq	#0,d0
	move.l	a1,d7
	sub.l	a0,d7
	subq.w	#1,d7
.next
	add.w	(A0)+,d0
	dbf	d7,.next
	move.w	d0,(A2)
	rts
	endc

checksum_6	macro
	marker	6,\1,\2,.sum\@

	i_leo1	\1,0
	i_lea	1,1		; constants for speed
	i_lea	2,2
	i_leo2	\2-\1,3		; counter
	i_lea	0,6		; 6 = checksum
	i_lea	.next\@,5		; loop
	
.next\@
	i_movef.w	0,4
	i_add.w		4,6
	i_add.l		2,0
	i_sub.l		1,3
	i_bne		3,5
; here r6 = checksum
	i_leo2	\3,0
.sum\@	equ	*+4
	i_lea	1234,5
	i_lea	100,4
	i_sub	5,6
	i_leo2	$4e75,5
	i_lea	.fill\@,7
	i_bne	6,7
	i_bra	.end\@
.fill\@
	i_movet.w 5,0 
	i_add.l	2,0
	i_sub	1,4
	i_bne	4,7
.end\@
	endm


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 7 - Definatly Fatal CRC checksum (sub.b, rol.l) - quite quick

	ifd	mastering
do_checksum_7
	clr.l	d0
	move.l	a1,d7
	sub.l	a0,d7
	subq.l	#1,d7
.next
	sub.b	(A0)+,d0
	rol.l	#1,d0
	dbf	d7,.next
	move.w	d0,(A2)
	rts
	endc

checksum_7	macro
	marker	7,\1,\2,.sum\@
	move.l	#(\1)/2,a0    			; BRACKETS!!!
	add.l	a0,a0
	clr.l	d0
	move.w	#((\2)-(\1))-1,d7    		; !!!
.next\@
	sub.b	(A0)+,d0
	rol.l	#1,d0
	dbf	d7,.next\@
.sum\@	equ	*+2
	sub.w	#5678,d0
	beq.s	.ok\@
	ifeq	checksum_crash
	move.l	#\4,d1
	jsr	maybe_error
	endc
; now do a fatal crash
	move.w	#$2700,sr  		; ints off
	move.l	#-1,sp	    		; impossible stack
	lea	$7fff8,A0
	move.l	.cd\@,(a0)
	move.l	.cd\@+4,4(a0)
	jmp	(a0)
.cd\@	movem.l	d0-d7/a0-a6,-(a0)      	; clear all chip memory
	bra.s	.cd\@	
		

.ok\@
	endm



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Run time encryption routines


; 1 - a fairly simple one

	ifd	mastering
do_run_time_1
	move.l	a1,d0
	sub.l	a0,d0
	lsr.l	#1,d0
	subq.w	#1,d0
	move.w	(A0)+,d1
.next
	eor.w	d1,(a0)
	move.w	(A0)+,d1
	dbf	d0,.next
	rts
	endc

start_run_time_1	macro	lable
	rmarker	1,run_start\1,run_end\1,0
	lea	run_start\1,a0
	move.l	#(run_end\1-run_start\1)/2-1,d0
	move.w	(A0)+,d1
nextxx\1
	move.w	(A0),d2
	eor.w	d1,(a0)+
	move.w	d2,d1
	dbf	d0,nextxx\1
	jsr	\2
run_start\1
	endm

end_run_time_1	macro	lable
	lea	run_start\1,a0
	move.l	#(run_end\1-run_start\1)/2-1,d0
	move.w	(A0)+,d1
run_end\1
.next\@
	eor.w	d1,(a0)
	move.w	(a0)+,d1
	dbf	d0,.next\@
	endm
	
;;		
; 2 - bit more complex, with no instruction sequence common with 1

	ifd	mastering
do_run_time_2
	move.l	a1,d0
	sub.l	a0,d0
	lsr.l	#1,d0
	subq.w	#1,d0
	move.w	(A0)+,d1
.next
	eor.w	d0,(a0)
	eor.w	d1,(a0)
	move.w	(A0)+,d1
	dbf	d0,.next
	rts
	endc

start_run_time_2	macro	lable
	rmarker	2,run_start\1,run_end\1,0
	lea	run_start\1,a0
	move.w	(A0)+,d1
	move.w	#(run_end\1-run_start\1)/2-1,d0
nextxx\1
	move.w	(A0),d2
	eor.w	d1,(a0)
	eor.w	d0,(a0)+
	move.w	d2,d1
	addq.w	#1,d2 		; just to diff it
	dbf	d0,nextxx\1
	jsr	\2
run_start\1
	endm

end_run_time_2	macro	lable
	lea	run_start\1,a0
	move.l	#(run_end\1-run_start\1)/2-1,d0
	move.w	(A0)+,d1
run_end\1
.next\@
	eor.w	d1,(A0)
	eor.w	d0,(a0)
	move.w	(a0)+,d1
	addq.w	#1,d2		; diff it
	dbf	d0,.next\@
	endm

;;		
; 3 - encode one bit of program with another (equivalent to checksumming)
; make sure you have enough unchanging code in the source to cover dest!
; Ideally the source should be a protection routine and the dest
; should contain checksums etc

	ifd	mastering
do_run_time_3
	move.l	a1,d0
	sub.l	a0,d0
	lsr.l	#1,d0
	subq.w	#1,d0
.next
	move.w	(a2)+,d1
	add.w	d1,(a0)+
	dbf	d0,.next
	rts
	endc

start_run_time_3	macro	lable,source
	rmarker	3,run_start\1,run_end\1,\2
	lea	run_start\1(pc),a0
	lea	(\2)-160,a1
	lea	160(A1),a1
	move.w	#(run_end\1-run_start\1)/2-1,d0
.nextxx\1
	move.w	(A1)+,d2
	sub.w	d2,(a0)+
	dbf	d0,.nextxx\1
	jsr	\3
run_start\1
	endm

end_run_time_3	macro	lable,source
	lea	run_start\1,a0
	lea	\2,a1
	move.l	#(run_end\1-run_start\1)/2-1,d0
run_end\1
.next\@
	move.w	(a1)+,d2
	add.w	d2,(a0)+
	dbf	d0,.next\@
	endm



;;		
; 4 - rather more complex, only one encoding routine used
; end_run_time_4 must be before an rts

	ifd	mastering
do_run_time_4
	move.l	a1,d0
	sub.l	a0,d0
	lsr.l	#1,d0
	subq.w	#2,d0
.next
	neg.w	(a0)+
	dbf	d0,.next
	rts
	endc

start_run_time_4	macro	lable
	rmarker	4,run_start\1,run_end\1,0
	bsr	run_code\1
	jsr	\2
run_start\1
	endm

end_run_time_4	macro	lable
run_end\1
run_code\1
	lea	run_start\1,a0
	move.l	#(run_end\1-run_start\1)/2-2,d0
.next\@
	neg.w	(a0)+
	dbf	d0,.next\@
	endm



;;		
; 5 - variation on 4, disguised
; end_run_time_5 must be before an rts

	ifd	mastering
do_run_time_5
	move.l	a1,d0
	sub.l	a0,d0
	lsr.l	#1,d0
	subq.w	#2,d0
.next
	move.w	(a0),d1
	ror.w	#8,d1
	move.w	d1,(A0)+
	dbf	d0,.next
	rts
	endc

start_run_time_5	macro	lable
	rmarker	5,run_start\1,run_end\1,0
	bsr	run_code\1
	jsr	\2
run_start\1
	endm

end_run_time_5	macro	lable
run_end\1
run_code\1
	move.l	#(run_end\1-run_start\1)/2-2,d0
	lea	run_start\1,a0
.next\@
	move.w	(a0),d1
	ror.w	#8,d1
	move.w	d1,(A0)+
	dbf	d0,.next\@
	endm



; **************************************************
; REMEMBER to put new routines in the CHECKSUM_TABLE
***********************************************;;;;;;;;!!!!!!!
	
	
; find the list of checksums then 
; generate all the checksums from the list, deleting the list.
	ifd	mastering
	include	\dev\random.s
; a0 = start, d0 = length
generate_checksums
	move.l	a0,a5		; a5 = base
generate_some_checksums
	move.l	a0,a4
.seek
	cmp.l	#c_marker,(a4)+
	beq.s	.ok
	subq.l	#2,a4
	bra.s	.seek
.ok

	sub.l	#base,a5	; a5 = offset for addrs
.next_word
	move.w	(a4),d7			; cs type
	bmi.s	.done
	move.w	#$4e71,(a4)+
	move.l	(a4),a0	       		; start
	move.l	#$4e714e71,(a4)+
	move.l	(a4),a1		     	; end 
	move.l	#$4e714e71,(a4)+
	move.l	(a4),a2		  	; cs poke addr
	move.l	#$4e714e71,(a4)+

	add.l	a5,a0
	add.l	a5,a1
	add.l	a5,a2

	add.w	d7,d7
	add.w	d7,d7
	move.l	checksum_table-4(pc,d7.w),a3
	pushem
	jsr	(a3)
	popem
	bra	.next_word
.done
	rts


checksum_table
	dc.l	do_checksum_1
	dc.l	do_checksum_2
	dc.l	do_checksum_3
	dc.l	do_checksum_4
	dc.l	do_checksum_5
	dc.l	do_checksum_6
	dc.l	do_checksum_7
	

; a list of single word instructions that do not alter program flow

instruction_list
	moveq	#0,d0
	add.l	d0,d0
	addq.l	#1,d0
	mulu	d0,d0
	move.b	(A1),d0
	clr.w	d0			
	moveq	#0,d1
	add.l	d1,d1
	addq.l	#1,d1
	mulu	d1,d1
	move.b	(A1),d1
	clr.w	d1
	moveq	#0,d2
	add.l	d2,d2
	addq.l	#1,d2
	mulu	d2,d2
	move.b	(A1),d2
	clr.w	d2

instruction_list_end


; encode the bits fo the file that need run time encryption
; do it backwards, so can have nested encryption
; a0 = start, d0 = length
generate_run_times
	move.l	a0,a5		; a5 = base
generate_some_run_times

	move.l	a0,a4
; find the start of the RT list
.seek
	cmp.l	#r_marker,(a4)+
	beq.s	.ok
	subq.l	#2,a4
	bra.s	.seek
.ok
; now find the end, as we want to do them in reverse order
; records are 14 bytes long
.seek_end
	tst.w	(a4)
	bmi.s	.end
	lea	14(a4),a4
	bra.s	.seek_end
.end
	sub.l	#base,a5	; a5 = offset for addrs
.next
	move.l	-(a4),a2
	cmp.l	#r_marker,a2
	beq.s	.done
	clr.l	(a4)
	move.l	-(a4),a1
	clr.l	(a4)
	move.l	-(a4),a0
	clr.l	(a4)
	move.w	-(a4),d7
	clr.w	(a4)

	pushem
	add.l	a5,a0
	add.l	a5,a1
	add.l	a5,a2

	add.w	d7,d7
	add.w	d7,d7
	move.l	run_time_table-4(pc,d7.w),a5
	jsr	(a5)
	popem
	bra	.next
.done
	rts

run_time_table
	dc.l	do_run_time_1
	dc.l	do_run_time_2
	dc.l	do_run_time_3
	dc.l	do_run_time_4
	dc.l	do_run_time_5
	endc



	ifd	parasol_stars
	if	~using_checksums
; macros are redefined here if not using checksums

checksum_1	macro
		endm			
checksum_2	macro
		endm			
checksum_3	macro
		endm			
checksum_4	macro
		endm			
checksum_5	macro
		endm			
checksum_6	macro
		endm			
checksum_7	macro
		endm			
checksum_8	macro
		endm			
checksum_9	macro
		endm			
checksum_10	macro
		endm   

start_run_time_1	macro
			endm
start_run_time_2	macro
			endm
start_run_time_3	macro
			endm
start_run_time_4	macro
			endm
start_run_time_5	macro
			endm
start_run_time_6	macro
			endm
end_run_time_1		macro
			endm
end_run_time_2		macro
			endm
end_run_time_3		macro
			endm
end_run_time_4		macro
			endm
end_run_time_5		macro
			endm
end_run_time_6		macro
			endm

		endc
		endc			

; put this at the end of your code
; mark the end of the checksum gen table
	ifd	parasol_stars
	pushs
	section		last_cs,text cs_end_group
	dc.l	-1
	section		last_rt,text rt_end_group
	dc.l	-1
	pops
	endc



; end of CHECKSUM.S
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

