;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; COPYLOAD.S - Mick West. Jan 1991
; 
; Disk routines, using the Rob Northen Copylock DISKIO routine.
; My own silly file structure on disk, see MASTER.S
;
; (Should) work on ST and AMIGA
;
; all routines return with d1 = -ve if error
; or d1 = +ve length if ok
; *** not bombproof yet, do it sometime!
;
; you should	JSR  FLOPPY_INT    every vbl to turn the motor off

; If files are arranged contiguously on disk then you can speed up the 
; load by >30% by using a 512 byte buffer, simply equ USE_SECTOR_BUFFER to 1
; and have the label SECTOR_BUFFER point to a spare 512 bytes
; this helps as the last sector of one file will also be the first sector
; of the next file, hence the number of calls to diskio is reduced.

; set CACHE_FILES.b to cache in extra memory (defaults to ON)

	ifnd	use_sector_buffer
use_sector_buffer	equ	0
	endc


	ifnd	do_wind
do_wind	equ	1
	endc

	ifnd	use_rnc_unpack
use_rnc_unpack	equ	-0
	endc


	if	st
	include	\copylock\diskio.s 	; Atari ST version
	elseif
	include	\alock\diskio.s		; Amiga Version
	endc

copyload_start


dir_size	equ	6 		; if changed then change MASTER.S

		if	st
dir_sector	equ	10  		; Ditto
		elseif
dir_sector	equ	2 		; Ditto
		endc


	include	\dev\io_macros.s

; Load a file, input
;	a0 = filename
;	a1 = file destination
;	a2 = directory buffer (dir_size*512+10, say 2k)
;  output
;	d1 = length


; if dest is in fast mem then we will need to load it into chip mem first
; and then copy it down with the processor

load_file
	movem.l	d0/d2-d7/a0-a6,-(Sp)
	cmp.l	#$100000,a1			; is it in chip
	bgt.s	.fast	
	bsr	load_file_chip		     	; load as normal if so
	bra.s	.done
.fast
	move.l	a1,-(sp)
	move.l	chip_load_buffer,a1   		; must be big enought
	jsr	load_file_chip
	move.l	(sp)+,a1
	movem.l	d0/a0/a1,-(sp)
	move.l	d1,d0
	lsr.w	#2,d0				; will mybe do 4 bytes over
	move.l	chip_load_buffer,a0
.copy_file
	move.l	(a0)+,(a1)+	  		; copy down
	dbf	d0,.copy_file	  		; dbf assumes <256k
	movem.l	(sp)+,d0/a0/a1


.done
	if	use_rnc_unpack
	move.l	a1,a0
; check for RNC compression here
	clr.l	d0
	jsr	unpack		; unpack RNC
	tst.l	d0
	ble.s	.not_rnc
	move.l	d0,d1	  	; get new length
.not_rnc
	endc


	movem.l	(sp)+,d0/d2-d7/a0-a6
	
	rts

load_file_chip
	movem.l	d0/d2-d7/a0-a6,-(Sp)
	move.l	a0,_file_name
	move.l	a1,_file_dest
	move.l	a2,_directory
	add.l	#8,_directory
retry_load
	jsr	load_directory
	tst.l	d1
	bmi	.error
	jsr	find_file		; return d0 = start, d1 = len
	tst.l	d1
	bmi	.error
	jsr	load_it			; load to a0
	tst.l	d1
	bmi	.error
	move.l	_file_len,d1
	move.l	_file_Dest,a0
	cmp.l	#$34567893,(a0)+  	; run it
	bne.s	.done

;	if	st
	move.l	#-1,directory+8
	pushem				; save a0
	jsr	load_directory 		; reset to track 0 to reassure Rob
	popem				; restore a0
;	endc

	clr.w	flock

	ifd	no_ints
	if	no_ints
	di	#$ff0
	endc
	endc

	jsr	(a0)  			; check disk and decrypt
	ifd	no_ints
	if	no_ints
	ei
	endc
	endc
	move.w	#100,flock
.done
	movem.l	(sp)+,d0/d2-d7/a0-a6




	rts
.error
; some error has occured - so retry the load, loading the directory again
; (which should reassure the diskdrive as to which track it is on, hopefully)
	move.l	_directory,a6
	clr.l	-8(a6) 				; force directory reload
	move.w	#-1,buffer_number 		; ignore sector cache
	bra	retry_load
	

read	macro	sector,no,addr
	move.w	\1,d1
	move.w	\2,d2
	move.l	\3,a0
	jsr	read_sectors
	endm

; load the directory sectors, return -1 if error
load_directory
	move.l	_directory,a6
	cmp.l	#"DIRE",-8(a6) 		; is it there ?
	bne	.load_it
	bsr	dir_checksum	  	; is the checksum ok ?
	cmp.l	-4(a6),d7
	beq	.done
.load_it

	read	#dir_sector,#dir_size,_directory
	move.w	d0,d1
	neg.w	d1
	bmi	.error
	move.l	_directory,a6
	bsr	dir_checksum
	move.l	d7,-4(a6)
	move.l	#"DIRE",-8(a6)
.done
	clr.l	d1
	rts
.error
	moveq.l	#-1,d1
	rts


; read the directory, turning motor off
do_motor_off
	read	#dir_sector,#dir_size,_directory
	move.w	#dir_sector,d1
	move.w	#dir_size,d2
	move.l	#_directory,a0
	jsr	read_sectors_motor_off
	rts

dir_checksum
	move.l	a6,a5
	move.l	#0,d7
	move.w	#(dir_size*512)/4-1,d6
.next_check
	add.l	(a5)+,d7
	dbf	d6,.next_check
	rts


find_file   
	move.l	_directory,a0
	addq.l	#2,a0
	move.w	#dir_size*25-1,d7		; entries to check
.next_entry
	move.l	_file_name,a1
	move.w	#12-1,d6
.next_letter
	move.b	(a1)+,d0
	beq	.got_it
	move.b	(a0)+,d1
	ror.b	#4,d1
	cmp.b	d1,d0
	beq.s	.same_letter
	eor.b	#$20,d1
	cmp.b	d1,d0
	beq	.same_letter
	add.w	d6,a0			; Skip remaining chrs
	addq.l	#8,a0
	dbf	d7,.next_entry		; Wow! Such convolution!
	moveq.l	#-1,d0
	moveq.l	#-1,d1
	rts
.same_letter
	dbf	d6,.next_letter	     	; Improper nesting!!
.got_it
	add.w	d6,a0	   	       
	addq.l	#1,a0  		       
	move.l	(a0)+,d0
	eor.l	#$4e4e4e4e,d0		; Exclamation Mark!
	move.l	(a0)+,d1
	eor.l	#$4e4e4e4e,d1
	rts

; input d0 = start position, d1 = length
load_it
	move.l	d0,_file_start
	move.l	d1,_file_len
; work out the address of the first sector
	divu	#512,d0
	move.l	_directory,a0
	add.w	(a0),d0			; d0 = first sector
; Load the first sector
	move.w	d0,d1
	move.w	d0,d7
	moveq	#1,d2
	move.l	_file_dest,a0

	if	use_sector_buffer
	cmp.w	buffer_number,d1	; is it the one in the buffer?
	bne	.not_buffered
	pushem
	lea	sector_buffer,a1 	; copy it across
	move.w	#512/4-1,d7
.copy_from_buffer
	move.l	(A1)+,(a0)+
	dbf	d7,.copy_from_buffer
	popem
	bra	.got_first_sector
.not_buffered
	endc

	bsr	read_sectors

	if	use_sector_buffer
	pushem
	move.w	d1,buffer_number
	lea	sector_buffer,a1 	; copy it across
	move.w	#512/4-1,d7
.copy_to_buffer
	move.l	(A0)+,(a1)+
	dbf	d7,.copy_to_buffer
	popem
	endc

.got_first_sector
; move it down if not on a sector boundry
	move.l	_file_dest,a0
	lea	512(a0),a0
	move.l	_file_start,d0
	and.l	#$1ff,d0 		; d0 = bytes over sector boundry
	move.l	#512,d1
	sub.l	d0,d1 			 
	move.l	d1,d3			; d3 = bytes to move
	beq	.on_sector_boundry

	subq.l	#1,d1
	lea	-512(a0),a0
	move.l	a0,a1
	add.w	d0,a0
.copy_down
	move.b	(a0)+,(a1)+
	dbf	d1,.copy_down
	move.l	a1,a0			; a0 = where to load remainder 		

.on_sector_boundry
	move.w	d7,d1			; last sector loaded
	addq.w	#1,d1			; go to next sector
	move.l	_file_len,d2		; len of file
	sub.l	d3,d2			; get bytes remaining
	ble	.decrypt		; check if less than 512 bytes

.load_the_rest
; finally load any remaining sectors (will trash up to 510 bytes)
; needs d1 = sector to load from, d2 = bytes remaining
	add.l	#510,d2			; round up
	divu	#512,d2			; no of sectors
	beq	.decrypt		; QQ not sure about this
	bsr	read_sectors		; was  _motor_off, but not anymore


	if	use_sector_buffer
	pushem
	tst.w	d0
	bne.s	.dont_buffer_error
	subq.w	#1,d2
	add.w	d2,d1			; last sector number
	move.w	d1,buffer_number 	; store it
	mulu	#512,d2
	add.l	d2,a0			; addr of last sector
	lea	sector_buffer,a1 	; copy it across
	move.w	#512/4-1,d7
.copy_last_to_buffer
	move.l	(A0)+,(a1)+
	dbf	d7,.copy_last_to_buffer
.dont_buffer_error
	popem
	endc

	move.w	d0,d1
	neg.w	d1
	ext.l	d1
	bmi.s	.error


.decrypt
; now DECRYPT it (eor with len counter, also calc the CRC)

	move.l	_file_dest,a0
	move.l	_file_len,d7
	subq.l	#4,d7			; get orig len (less checksum) 
	moveq.l	#0,d0			; zero the checksum
.next_byte
	move.b	(a0),d1	     		; get byte
	eor.b	d7,d1		  	; decrypt byte
	eor.b	d1,d0		   	; do checksum
	ror.l	#1,d0			; CRC
	move.b	d1,(A0)+	 	; store byte
	subq.l	#1,d7
	bne.s	.next_byte

	tst.l	(a0)	     		; 1 in 4.2 billion chance of fuckup
	beq.s	.error
	cmp.l	(A0),d0			; check checksum
	bne.s	.error

	subq.l	#4,_file_len	   	; correct len
	move.l	_file_len,d1		; return len if ok
	rts
.error
	moveq.l	#-1,d1
	rts
	
; input d1 = sector,d2 = no of sectors, a0 = buffer
read_sectors
	clr.w	flock
	if	st
	moveq.l	#8,d0
	elseif
	moveq	#0,d0
	endc
	move.w	#$0000,d3

.retry	
	if	amiga
	move.l	the_lzw_buffer,a1      			; use screen as track buffer
	move.w	#%1000001000010000,$dff096	; enable dma
	endc
	pushem
	jsr	diskio
	if	amiga
	waitblit
	endc
	tst.w	d0
	beq.s	.ok
	move.w	#-1,d0 		; little pause before retry
.x	dbf	d0,.x
	popem
	
	bra.s	.retry
.ok
	popem

	cmp.l	#0,(a0)
	bne.s	.not_zero
.hang_up
	add.w	#1,d0
	move.w	d0,$dff180
	bra	.hang_up

.not_zero

	if	st&~do_wind	
; reselect the drive to let the motor turn itself off
	move.w	#$2700,sr		ints off 
	move.b	#$e,(flselect).w
	move.b	(flselect).w,d1
	and.b	#$f8,d1			select both drives
	move.b	d1,(flwrite).w		pass to ACIA
	move.w	#$2300,sr		ints on
	endc

; drive off after 2 seconds (under interrupts)
	clr.l	d0
	move.w	#100,flock
	rts

; ditto, turning the motor off (a two second delay)
read_sectors_motor_off
	clr.w	flock
	if	st
	moveq.l	#8,d0
	elseif
	moveq	#0,d0
	endc
	move.w	#$8000,d3
	if	amiga
	move.l	the_lzw_buffer,a1
	move.w	#%1000001000010000,$dff096	; enable dma
	endc
	jsr	diskio
	if	amiga
	waitblit
	endc
	rts

;	Turn Drive Motor Off (Call On Vblank)
flselect	equ	$ffff8800
flwrite		equ	$ffff8802
floppy_int
	tst.w	flock		; For turning disc motor off
	beq	.nope
	subq.w	#1,flock
	bne	.nope
	if 	st
	move.l	d1,-(sp)
	move.w	sr,-(sp)

	move.w	#$2700,sr	; ints off 
	move.b	#$e,(flselect).w
	move.b	(flselect).w,d1
	and.b	#$f8,d1
	or.b	#7,d1  		; deselect both drives
	move.b	d1,(flwrite).w	; pass to ACIA

	move.w	(sp)+,sr
	move.l	(sp)+,d1

	elseif
	lea	CIA_B,a5
;	move.b	#$fd,(a5) 	; %11111100
;	rept	10
;	nop
;	endr
;	move.b	#$e7,(a5)	; %11100111
;	rept	10
;	nop
;	endr

	BSET	#3,(a5)			; deselect internal drive
	rept	10
	nop
	endr
	BSET	#7,(a5)		   	; motor off
	rept	10
	nop
	endr
	BClR	#3,(a5)		 	; latch by selecting internal drive
	rept	10
	nop
	endr
	Bset	#3,(a5)		 	; deselect internal drive
	endc
;	clr.w	flock
	move.w	#50,flock		; repeat every second
.nope	
	rts



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; routines for implementing the ram disk buffer for floppy usage

; check for extra memory, if any then setup an empty ramdisk
check_memory
	pushem
	tst.b	cache_files
	beq	.done_it
	tst.b	memory_checked
	bne	.done_it
	st	memory_checked
	sf	memory_there
	lea	ram_area_list,a1
.next_area
	move.l	(a1)+,a0
	cmp.l	#-1,a0
	beq.s	.done_it
	if	amiga
	clr.l	0
	endc
	move.l	#$12345678,(a0)
	if	amiga
	cmp.l	#$12345678,0	      	; if mapped over bottom 512k then
	beq.s	.nope		      	; not there
	endc
	cmp.l	#$12345678,(a0)		; check if it remembered that
	bne	.nope
	move.l	#$ffff0000,(a0)	 	; and try with a different bit pattern
	cmp.l	#$ffff0000,(a0)
	beq.s	.got_it
.nope
	bra.s	.next_area
	

.got_it
	move.l	a0,ram_buffer_area
	st	memory_there
; could check if there is lots, but just assume 512k for now
	move.l	#$80000,memory_left
; ok, we have another 512k, so setup the empty ramdisk	   
	move.l	#-1,(a0)		; soch eash dollin!	
.done_it
	popem
	rts



; try to load a file from ram disk, returns d1 = len if ok, -1 if not
; checksums - as in MAYBE_LOAD_FILE - hence avoids corruption problems
; and robust - will always exit
do_ram_load
	pushem	d0/d2-d7/a0-a6	

	move.l	#-1,d1			; file not found (yet)

	move.l	ram_buffer_area,a2
	lea	-2(a2),a6		; a6 = last file
.next_file

; if a0 is odd or <= last one or >=top of ram then file not found
; Guarenteed to exit!
	move.l	a2,d0
	btst	#0,d0			; check if odd
	bne	.file_not_found
	cmp.l	a6,a2		 	; or LE the last one
	ble	.file_not_found
	ifne	st
	cmp.l	#$400000,a2	 	; max mem on ST (4 meg)
	elseif
	cmp.l	#$c80000,a2 		; 512k expansion on amiga
	endc
	bge	.file_not_found
	move.l	a2,a6

	move.l	(a2)+,a3 		; a3 = this file start
	cmp.l	#-1,a3
	beq	.file_not_found
	move.l	(a2)+,d0		; d0 = length
	lea	4(a2),a4		; a4 = name address
	move.l	a0,a5
.check_name
	move.b	(A5)+,d2
	beq	.file_found
	cmp.b	(a4)+,d2
	beq	.check_name
	move.l	(a2),a2
	bra	.next_file
.file_found
	move.l	d0,d1			; will return len if checksum ok
	move.l	d0,d7
	addq.l	#1,d7
	bclr	#0,d7
	lsr.l	#1,d7
	clr.l	d0  			; ready CRC checksum
.copy_file
	move.w	(a3)+,d6
	eor.w	d6,d0			; calc CRC checksum
	ror.l	#7,d0
	move.w	d6,(a1)+
	subq.l	#1,d7
	bne	.copy_file
	cmp.w	(a3),d0	     		; check checksum
	beq.s	.done
	move.l	#-1,d1			; checksum failed
.done

.file_not_found
	popem	d0/d2-d7/a0-a6
	rts


maybe_load_file
	movem.l	d0/d2-d7/a0-a6,-(Sp)
	bsr	check_memory
	tst.b	memory_there
	beq	.nope
	bsr	do_ram_load
	tst.l	d1
	bpl	.loaded
.nope
	bsr	load_file
	tst.b	memory_there 		; if no extra memory
	beq	.loaded		 	; then dont cache
	tst.b	cache_files    		; if cache flag turned off
	beq	.loaded		  	; then dont cache
; now we have  a0 = filename, a1 = where it is, d1 = length
; so maybe make an entry in the ram buffer
; first check the length
	move.l	d1,d2
	add.l	#3*4+14,d2		; add max size of header
	cmp.l	memory_left,d2
	bgt	.loaded
; right, there is enough memory
	sub.l	d2,memory_left		; grab memory
; now find the last file in the ram buffer
	move.l	ram_buffer_area,a2
.find_end
	tst.l	(A2)
	bmi.s	.got_end
	move.l	8(A2),a2
	bra.s	.find_end
.got_end
; now a2 = the next entry in the ram disk, so set up the header
	move.l	d1,4(a2)	; length is known
	lea	12(A2),a3	; where the name goes
.copy_name
	move.b	(A0)+,(A3)+
	bne	.copy_name
	move.l	a3,d0
	addq.l	#1,d0
	bclr	#0,d0
	move.l	d0,a3
	move.l	a3,(a2)		; now got the start
; now copy the file into the ram disk
	move.l	d1,d7
	addq.l	#1,d7
	bclr	#0,d7
	lsr.l	#1,d7
	clr.l	d0 			; ready CRC checksum
.copy_file
	move.w	(A1)+,d6
	eor.w	d6,d0	      		; calc CRC checksum
	ror.l	#7,d0
	move.w	d6,(a3)+ 
	subq.l	#1,d7
	bne	.copy_file
	move.w	d0,(a3)+		; set checksum
	move.l	#-1,(A3)		; set new null entry
	move.l	a3,8(a2)		; and point this one to it 
		
	
.loaded
	movem.l	(sp)+,d0/d2-d7/a0-a6
	rts	

copyload_end	


	dc.l	0

memory_checked	dc.b	0
memory_there	dc.b	0
memory_left	dc.l	$80000
_file_name	dc.l	0
_file_dest	dc.l	0
_directory	dc.l	0
_file_start	dc.l	0
_file_len	dc.l	0

ram_buffer_area	dc.l	0

; list of possible start point for ram expansion
ram_area_list
	dc.l	$080000		; 512k extended chip memory (for a500+ and ST)
	if	amiga
	dc.l	$c00000	   	; 512k a500 internal expansion memory
	dc.l	$200000		; primary 8 meg auto-config space 
	dc.l	$300000		; and all it's one meg boundries
	dc.l	$400000		;
	dc.l	$500000		;
	dc.l	$600000		;
	dc.l	$700000		;
	dc.l	$800000		;
	dc.l	$900000		;
	endc
	dc.l	-1

flock	dc.w	0



buffer_number	dc.w	-1		; number of sector buffered

cache_files	dc.b	-1		; if set then cache files

		even
; end of COPYLOAD.S
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
