;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; MASTER.S - Mick West 1991
; reads the files from ram disk
; Read Files, stonk them together and write them out to sectors
;  together with a directory.  Written initially for Darkman.

; also put a boot program on the boot sector to load the first file in
; the list, it is loaded to the addr in $432 (_membot) and run from there
; so should be relocatable (uses the Floprd so ignores boot sector config)

; needs a formated double sided disk in internal to write to

; these equates specify the layout of the disk
; the sectors are LOGICAL, so will depend on the format of the disk
; remember that sector 6 of COPYLOCK disks is protected and must be unused.

; if ram_pass = 2 then add the files to those already there

	ifnd	cs_group
cs_group	group
	endc


		if	st
directory_sector	equ	10 	; where to put the direcory (was 7)
directory_len		equ	6 	; 25 files per sector
data_sector		equ	22 	; where to put the main files
colour0			equ	$ff8240
		elseif
directory_sector	equ	2  	; where to put the direcory
directory_len		equ	6 	; 25 files per sector
data_sector		equ	22 	; miss more on amiga
colour0		equ	$dff180
		endc

use_io	equ	1   		; do we use the DISKIO ?    - yes

	if	st
fn	equ	8
	elseif
fn	equ	0
	endc


; directory format is

; SECTOR.W		- no of the first sector (7 for st, 22 for amiga)
;  then lots of files
; FILENAME.EXT		- 12 bytes, not zero terminated
; START.L		- 1 LW, offset in bytes of start of file
; LEN.L			- len in bytes of file.
;  (total 20 bytes for each directory entry)
;  the directory is mildly engrypted for additional protection :-
;  the chars in the filename have their nibbles swapped
;  the start and len are EOR.Led with $4e4e4e4e

	lea	the_stack,sp
	bra	start_batch

	if	st
	include	\copylock\diskio.s	; ST disk read/write routines
	elseif
	include	\alock\diskio.s		; AMIGA disk read/write routines
	endc	


pushem	macro
	movem.l	d0-d7/a0-a6,-(sp)
	endm

popem	macro
	movem.l	(sp)+,d0-d7/a0-a6
	endm

	include	c:\dev\print.s
	include	checksum.s
	include	encrypt1.s


real	dc.l	0

dos	macro	call,tidy
	move.w	\1,-(sp)
	ifeq	use_io
	trap	#1
	endc
	add.l	\2,sp
	tst.l	d0
	endm

print	macro	string.l	
	ifeq	run
	move.l	\1,-(sp)
	dos	#9,#6
	endc
	endm


start_batch:
	
	ifne	amiga
	move.w	#%1000001000010000,$dff096	; enable dma
	endc


	move.l	sp,old_stack
	print	#start_mess
	move.l	#bbuffer,file_dest
	if	ram_pass=1		  	; pass 1 so
	clr.l	total_len   			; no files done yet
	move.w	#data_sector,directory	      	; store start of directory
	move.l	#directory+2,dir_addr 		; addr in directory
	move.l	#-1,directory+2			; zero directory
	clr.b	first_done			; none done yet
	elseif
	st	first_done    			; pass 2, at least one done
; we need to load the directory etc here
	moveq	#fn,d0
	moveq	#directory_sector,d1
	move.w	#directory_len,d2 
	move.w	#0+$8000,d3
	move.l	#directory,a0
	lea	track_buffer,a1
	jsr	diskio	      				; load dir
; now search through for the last one
	lea	directory+2,a0
.x	cmp.l	#-1,20(a0)
	beq	.gotit
	lea	20(A0),a0
	bra	.x
.gotit
	move.l	12(a0),d0
	eor.l	#$4e4e4e4e,d0
	move.l	16(a0),d1
	eor.l	#$4e4e4e4e,d1
	add.l	d1,d0
	move.l	d0,total_len	 	; set total_len (pos on disk)
	lea	20(a0),a0
	move.l	a0,dir_addr  		; set pos in directory
	endc

	move.l	#_ram_start,a0 		; first file in ramdisk
main_loop
; we simply need to go through the files in order on the ram disk
; load them and get the file names and length (file in upper case)

	tst.b	first_done
	beq	.already

	move.l	(A0),a1			; get source
	cmp.l	#-1,a1
	beq	exit
	move.l	4(A0),d0		; length

	addq.l	#1,d0			; even it up
	bclr	#0,d0

	move.l	d0,file_len
	lea	12(A0),a2			; a2 = filename
	move.l	8(a0),a0			; next file
.already
	move.l	a0,-(sp)			; save next file

	jsr	get_file
	jsr	print_stuff

	move.l	(sp)+,a0				; restore pos in files
	bra	main_loop				; and loop


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; get the file from ram disk and make the directory entries
get_file
	move.l	#bbuffer,file_dest
	tst.b	first_done
	beq	.got_first   
	
; now copy the filename into file_name and copy the file to file dest
	lea	file_name,a3
.next_letter
	move.b	(A2)+,(a3)+
	bne	.next_letter

; right, the file is at a1,d0, we first need to load the sector it
; is on and to copy the file to the correct position after it

	move.l	total_len,d1
	divu	#512,d1
	move.l	d1,d2
	add.w	#data_sector,d1		; d1 = sector 
	swap	d2	  		; d2 = offset in sector
; reet, load the sector
	pushem
	lea	bbuffer,a0
	moveq	#fn,d0
	move.w	#1,d2
	move.w	#0,d3
	lea	track_buffer,a1
	jsr	diskio	 		; load last sect of prev file
	popem

	lea	bbuffer,a0
	add.w	d2,a0	  		; offset into sector
	move.l	a0,file_dest		; where the file goes

; now copy the file itself
	move.l	file_dest,a3
.next_byte
	move.b	(a1)+,(a3)+
	subq.l	#1,d0
	bne	.next_byte
	bra	.done_first

.got_first
	st	first_done
; we now have the main prog at FILE_DEST of len FILE_LEN
; any encrypting should be done here, returning the file at the
; same place with modified length

	bsr	do_protection

	move.l	file_len,first_len		; set prog len in boot sector
	bra	.encrypted
.done_first
; !!!!!!!!!!!!!!!!!!!  NEW ENCRYPTION & CHECKSUM
; each file is mildly encrypted by sucessive EORing bytes
; with the low byte of the loop counter (orig len)
; also the len is increase by 4 bytes which is a CRC (eor.b, ror.l)
; of the UNENCRYPTED data
; - see \DEV\COPYLOAD.S for decrypt/checking routine
	pushem
	move.l	file_dest,a0	  	; assumes even start !
	move.l	file_len,d7  		; and even length !
	moveq.l	#0,d0			; zero the checksum
.enc_next_byte
	move.b	(a0),d1	     		; get byte
	eor.b	d1,d0		   	; do checksum
	ror.l	#1,d0			; CRC
	eor.b	d7,d1		  	; encrypt byte
	move.b	d1,(A0)+	 	; store byte
	subq.l	#1,d7
	bne.s	.enc_next_byte
	move.l	d0,(a0)			; store checksum
	addq.l	#4,file_len		; mod the len to include this
	popem			
.encrypted
; we need to make an entry in the directory here
; of filename.ext,start offset,len.
	move.l	dir_addr,a6

	lea	file_name,a5
	move.w	#12-1,d7
.copy	move.b	(a5)+,d0
	ror.b	#4,d0
	move.b	d0,(a6)+
	dbf	d7,.copy
	move.l	total_len,d0
	eor.l	#$4e4e4e4e,d0
	move.l	d0,(a6)+
	move.l	file_len,d0
	eor.l	#$4e4e4e4e,d0
	move.l	d0,(a6)+
	move.l	#-1,(a6)
	move.l	a6,dir_addr
	
; finally we must write the file out to disk again
	move.l	total_len,d1
	divu	#512,d1
	add.w	#data_sector,d1		; d1 = sector 
; reet, save out the file
.rewrite
	pushem
	lea	bbuffer,a0		; include the partial sector we loaded
	tst.b	use_file_dest
	beq.s	.dont_use_file_dest
	move.l	file_dest,a0
	sf	use_file_dest
.dont_use_file_dest
	moveq	#fn,d0
	move.l	file_len,d2
	divu	#512,d2
	add.w	#2,d2			; more for luck!
	move.w	#1,d3 			; write file
	lea	track_buffer,a1
	jsr	diskio			; write the rest of the file
	tst.w	d0
	beq	.no_error
	moveq	#-1,d0
.err	move.w	d0,colour0    		; show there was an error
	dbf	d0,.err
	bra.s	.err
	popem
	bra	.rewrite

.no_error
	popem
	ifne	st
	move.w	file_len+2,$ff8240	; some indication of activity
	else
	move.w	file_len+2,$dff180
	endc
	

; advance the total len by an even amount
	move.l	file_len,d0
	add.l	#1,d0
	and.l	#$fffffe,d0
	add.l	d0,total_len
;	add.l	d0,file_dest
	rts


print_stuff
	print	#save_mess
	print	#file_name
	print	#space
; print filename,size,total size
	pushem
	print	#tab
	move.l	file_len,d0
	move.w	#7,d7
	move.l	#pprint,a0
	jsr	print_using
	print	#tab
	move.l	total_len,d0
	move.w	#7,d7
	move.l	#pprint,a0
	jsr	print_using
	popem
	print	#cr
	rts


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 
; Exit the program, writing the disk first
;

exit	
	move.l	old_stack,sp
	print	#directory_mess


	moveq	#fn,d0
	moveq	#directory_sector,d1
	move.w	#directory_len,d2 
	move.w	#1+$8000,d3
	move.l	#directory,a0
	lea	track_buffer,a1
	jsr	diskio			; write directory

	if	ram_pass=1
	jsr	write_boot   		; only first time round
	endc

	clr.w	d0
.x
	trap	#0
	eor.w	#$0f0,d0
	if	st
	move.w	d0,$ff8240
	elseif
	move.w	d0,$dff180
	endc
	bra	.x



pprint
	pushem
	move.b	d0,dat
	print	#dat
	popem
	rts


; input d0 = no of sectors
write_boot
	if	st
	lea	dev(pc),a0	; get device
	move.w	#0,(a0)

	sub.l	a5,a5		; Clear A5 (wow!)

	moveq	#8,d5		; XBIOS _FLOPRd code in d5
	bsr	read_write	; read boot sector

	move.l  #batch,A0	; a0 = disk buffer

	move.l	#$60380666,d0
	move.l d0,(A0)     	; BRA.S $3a
	move.w	#$7fff,4(a0)	; immune to some viruses

	lea	6(a0),a1  	; clear the rest of the buffer
	move.w	#300,d0
.kill	clr.w	(a1)+
	dbf	d0,.kill

	lea $3a(a0),a1          	; where to put it
	lea mystart(pc),a0        	; my code
	move.w #myend-mystart-1,d0 	; length
mymove:
	move.b (A0)+,(A1)+    	; move code into boot sector
	dbf d0,mymove 

	clr.w 	d1
	move.l #batch,A0
	move.w #254,d0        	; add 255 words
nextadd:
	add.w (A0)+,d1
	dbf d0,nextadd
	move.w #$1234,d0	; $1234-sum of first 255 words
	sub.w d1,d0		; is the new checksum
	move.w d0,(A0)       	; insert checksum as last word

	moveq	#9,d5		; XBIOS _FLOPWR
	bsr	read_write	; write bootsector
	tst.w	d0
	bmi	.error
	bra	.ok	
.error
   	print	#error_retry
	bra	write_boot

.ok

	elseif


; poke the len into the loader
	move.l	first_len,d0
	add.l	#511,d0
	and.l	#$fffe00,d0		; round up to 512 byte boundry
	move.l	d0,amiga_len

	lea	bbuffer,a1

	move.w	#"DO",(A1)+
	move.b	#"S",(A1)+
	clr.b	(A1)+

	clr.l	(A1)+			; clear checksum
	move.l	#$370,(A1)+		; root block number.
	lea	amiga_boot_program,a0
	move.w	#amiga_boot_prog_end-amiga_boot_program-1,d0
.x	move.b	(a0)+,(A1)+
	dbf	d0,.x
; now calculate the checksum (from "Amiga DD inside and out", p101)
	lea	bbuffer,a0
	lea	4(a0),a1
	clr.l	(A1)
	move.w	#$ff,d1
	moveq	#0,d0
.loop	add.l	(A0)+,d0
	bcc	.jump
	addq.l	#1,d0  
.jump	
	dbf	d1,.loop
	not.l	d0
	move.l	d0,(A1)
; finally write the boot sector using diskio			
	moveq	#fn,d0
	moveq	#0,d1
	moveq	#2,d2
	moveq	#1,d3
	move.l	#bbuffer,a0
	move.l	#track_buffer,a1
	move.w	#%1000001000010000,$dff096	; enable dma
	jsr	diskio		 	; write amiga boot sector
	
	rts

where	equ	$60000		; a (hopefully) safe place

amiga_boot_program
; first turn off copper dma and display dma and set colours to black
	move.w	#%0000000110000000,$dff096
	lea	$dff180,a0
	move.w	#32-1,d0
.c	move.w	#0,(a0)+
	dbf	d0,.c

	move.l	a1,-(sp)    			; save disk info block addr
	move.l	4,a6
	
	move.l	#200000,d0			; alocate 200k
	move.l	#2,d1				; of chipmem
	jsr	-198(a6)
	move.l	d0,a5	 			; get addr
	move.l	(sp)+,a1   			; restore disk block info

	move.l	a5,-(sp)

	move.w	#2,$1c(a1)
	move.l	a5,$28(a1)			; where to load
amiga_len	equ	*+2
	move.l	#160*512,$24(A1)		; 80k length
	move.l	#22*512,$2c(a1)	   		; offset
	jsr	-$1c8(a6)
	move.w	#4,$1c(a1)
	jsr	-$1c8(A6)

; need to go supervisior mode first
	move.l	4,a6
	jsr	-150(a6)
	move.w	#$2700,sr

	bsr	cache_off

	move.l	(sp)+,a6  		; NEEDS TO BE A6 !!!!
	jmp	(a6)

	include	\dev\cacheoff.s

; Bung some text here
	
	dc.b	"                                                  "
	dc.b	"          This is a BOOT SECTOR and NOT a virus, "
	dc.b	"if you do anything to it then PARASOL STARS will not work "
	dc.b	"                                                        "


	rept	20
	dc.b	"PARASOL STARS - OCEAN - "
	endr
 
amiga_boot_prog_end
	even
	endc

	rts


; This is a routine to read or write the boot sector
; d5 is either 8 or 9 to read or write, or 19 to verify

read_write
	moveq	#fn,d0
	moveq	#0,d1
	moveq	#0,d3
	cmp.w	#8,d5
	beq	.read
	move	#1+$8000,d3
.read	move.l	#batch,a0
	lea	track_buffer,a1
	moveq	#1,d2			; ONLY ONE SECTOR !!!!!!!
	jsr	diskio

	rts			; return from reading or writing


dev	dc.w	0

first_done	dc.w	0

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Boot loader - will load the first program in the list and run it
; assumes that the program is relocatable
mystart
	lea	$ff8240,a0		; blank colours to black
	moveq	#8-1,d0
	moveq	#0,d1
.x	move.l	d1,(a0)+
	dbf	d0,.x

	move.l	$432,a6			; where to load to (MEMBOT)
	move.l	a6,-(sp)
	move.w	#0,d6			; track
	move.w	#data_sector,d7		; start sector
	move.w	#0,d5			; d5 = side
	bsr	read_track
	clr.w	d7

first_len	equ	*+2

	move.l	#12345678,d3		; length of the first program
	divu	#5120,d3
	add.w	#1,d3
.next
	eor.w	#1,d5		; change sides
	bne	.no_step
	add.w	#1,d6		; next track if back to side zero
.no_step
	bsr	read_track	; read a side (5120 bytes)
	dbf	d3,.next
.none
	move.w	#$2700,sr
	move.l	$432,a6			; where to load to (MEMBOT)
	rts		; Execute! (SR=$2700,Stack undefined)

read_track
	move.w	#10,d4
	sub.w	d7,d4
	movem.l	d0-d7/a0-a6,-(sp)
	addq.w	#1,d7			; sectors are 1-10 not 0-9
	move.w	d4,-(sp)
	move.w	d5,-(sp)
	move.w	d6,-(sp)
	move.w	d7,-(sp)
	move.w	#0,-(sp)
	clr.l	-(Sp)
	move.l	a6,-(Sp)
	move.w	#8,-(sp)
	trap	#14
	lea	20(sp),sp
	movem.l	(sp)+,d0-d7/a0-a6
	mulu	#512,d4
	add.l	d4,a6
	rts
	
; can put a message here if you want
	dc.b	"                                    "
	dc.b	"This is a boot sector and NOT a virus!"
	dc.b	" If you remove this then PARASOL STARS will NOT WORK "
	dc.b	"                                    "
	even

myend
; and ends here


; given file at FILE_DEST of len FILE_LEN, do any encryption,checksums etc
do_protection
	pushem
	bsr	do_checksums		; checksums will not alter length
; any other additional routines here
	if	encrypt_program
	jsr	encrypt_file  		; but this will!
	endc
	popem

; maybe just run it without bothering with other files
	ifne	run
	move.l	file_dest,a6
	jmp	(a6)
	endc

	rts

; at the start of the program there will be a list of areas that need
; checksumming or run-time encrypting.
; do these and then delete the list.
; remember not to include variables or self modifiying code in such areas
do_checksums
	move.l	file_dest,a0
	move.l	file_len,d0
	jsr	generate_checksums
	move.l	file_dest,a0
	move.l	file_len,d0
	jsr	generate_run_times
	rts

dat	dc.b	"x",0

file_name	dc.b	"DM.PRG",0,"             "


		dc.b	"0123456789012345678901234567890123456789"
start_mess	dc.b	27,"E"
		dc.b	"Mick's Magical Mastering Machine!",13,10,13,10
		dc.b	"Thinking about it ....",13,10,13,10,0
save_mess	dc.b	"Adding ",0

data_mess	dc.b	"Writing Data sectors",13,10,0
directory_mess	dc.b	"Writing Directory & Boot sector",13,10,0
well_mess	dc.b	13,10," All seems to be well",13,10
		dc.b	" As far as I can tell",13,10,0
error_retry	dc.b	" ERROR - I am retrying",13,10,0
cr		dc.b	13,10,0
tab		dc.b	9,0
space		dc.b	32,0
		even

name		ds.l	1	; addr of start of name
file_handle	ds.w	1
old_stack	ds.l	1
total_len	ds.l	1
file_dest	ds.l	1
dir_addr	ds.l	1


		list
file_len	dc.l	filesize('\ps\ps.prg')


		org	250*1024

; this is where bbuffer must point to
		incbin	\ps\ps.prg	

		rsset	150*1024
		rs.b	100*1024  	; space to expand into (max 100k)
bbuffer		rs.b	150*1024	; (at 250) where the raw program is

; here we are at 350*1024

		rs.b	2000
the_stack	rs.b	0
directory	rs.b	512*directory_len

		even
end		rs.b	0
track_buffer	rs.b	20000
batch		rs.b	10000

		nolist
