;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; OBBLE.S - An obscure interpreted language for protection obsfuscation
;
; very simple - loads of instructions to add
; and local stack to implement
;
; Have conditional instructions effect immediatly
;
; Hmm - May be best to have a risc and then use lots of macros
; that would be VERY difficult for the hacker to follow

; Remember to checksum the interpreter, as they could prevent writes
; to certain ranges !!!!!!!!!!!!!! This is important
; - could split up Macros and Code, so can have code in protected place

; Routines could either have their own stack, or use the normal one
; if doing a speedy single run of the entire program.

; Just need 7 instructions
;	LEA	reg,#number (or address)
;	MOVE	reg1,reg2
;	SUB	reg1,reg2
;	BNE	reg1,reg2	(if reg1<>0 then jmp to reg2)
;	END			(will return if sommat on stack, else stop)
;	MOVEF	(reg),reg
;	MOVET	reg,(reg)
;	JSR	reg		call a normal 68000 routine


; Instruction format could be (bytes) :

;	instruction, control byte (size, etc), r1,r2
;followed by a word or a long if required (data or addr)
;	END	will be two bytes (zero control byte)

; All blocks can be built from this
; Programs will be slower, but miles more complex.

; The defining macros can be sized as .W or .L

; Will need one, or probably two, temporary registers.

;st	equ	-1
;	include	d:\dev\kernal.s

;main:
;	lea	ob_program,a0
;	lea	ob_struct,a6
;
;	jsr	ob_run
;
;	lea	main,a0
;	move.w	#100-1,d7
;	clr.w	d0
;.next
;	add.w	(a0)+,d0
;	dbf	d7,.next
;
;	move.w	2+6*4(a6),d1  		; get r6
;
;	illegal
;my_vbl
;	ret

	rsreset
ob_r0		rs.l	1
ob_r1		rs.l	1
ob_r2		rs.l	1
ob_r3		rs.l	1
ob_r4		rs.l	1
ob_r5		rs.l	1
ob_r6		rs.l	1
ob_r7		rs.l	1
ob_r8		rs.l	1
ob_r9		rs.l	1
ob_r10		rs.l	1
ob_r11		rs.l	1
ob_r12		rs.l	1
ob_r13		rs.l	1
ob_r14		rs.l	1
ob_sp		rs.l	1
ob_pc		rs.l	1
ob_stack	rs.l	16
ob_len		rs.b	0



; given that there is an obble program in the code ahead, then
; run it and return to the address after it
; registers are preserved
ob_inline
	pushem
	move.l	15*4(sp),a0
	lea	ob_struct,a6
	bsr	ob_run
	move.l	ob_pc(a6),a0
	addq.l	#2,a0
	move.l	a0,15*4(sp)
	popem
	ret

; given a6 = language structure, a0 = program, run that program
ob_start
	move.l	a0,ob_pc(a6)
	ret

; given a6 = language structure, interpret a single instruction
ob_one
	pushem
	move.l	ob_pc(a6),a5
	moveq.l	#0,d0
	moveq.l	#0,d1
	move.b	(a5)+,d0
	move.l	ob_table(pc,d0.w),a4
	move.b	(a5)+,d1
	jsr	(a4)
	move.l	a5,ob_pc(a6)	
	popem
	ret


; run a task until finished
ob_run
	pushem
	move.l	a0,a5
.next
	move.l	a5,a3
	moveq.l	#0,d0
	moveq.l	#0,d1
	move.b	(a5)+,d0
	move.l	ob_table(pc,d0.w),a4
	move.b	(a5)+,d1
	jsr	(a4)
	move.l	a5,ob_pc(a6)	
	cmp.l	a5,a3
	bne	.next
	popem
	ret

; program command numbers
; in mults of 4

ob_end		equ	0		; end program, do not advance PC
ob_load		equ	4		; #data, reg
ob_move		equ	8		; #data, reg
ob_sub		equ	12		; addr
ob_bne		equ	16		; reg, addr
ob_movef	equ	20
ob_movet	equ	24
ob_jsr		equ	28		; call machine code routine
					; must preserve registers

ob_table
	dc.l	dob_end
	dc.l	dob_load
	dc.l	dob_move
	dc.l	dob_sub
	dc.l	dob_bne
	dc.l	dob_movef
	dc.l	dob_movet
	dc.l	dob_jsr


dob_end
;	add.w	#1,$ff8240
	subq.l	#2,a5
	ret

; load a long - reg is in size word
dob_load
	move.l	(a5)+,(a6,d1.w)
	ret

dob_move
	bne.s	.long
	move.b	(a5)+,d0
	move.b	(a5)+,d1
	move.w	2(a6,d0.w),2(a6,d1.w)
	ret
.long
	move.b	(a5)+,d0
	move.b	(a5)+,d1
	move.l	(a6,d0.w),(a6,d1.w)
	ret

dob_sub
	bne.s	.long
	move.b	(a5)+,d0
	move.b	(a5)+,d1
	move.w	2(a6,d0.w),d0
	sub.w	d0,2(a6,d1.w)
	ret
.long
	move.b	(a5)+,d0
	move.b	(a5)+,d1
	move.l	(a6,d0.w),d0
	sub.l	d0,(a6,d1.w)
	ret


; bne rx,ry
; unsized
dob_bne
	move.b	(a5)+,d0
	move.b	(a5)+,d1
	tst.l	(a6,d0.w)
	beq.s	.stay
	move.l	(a6,d1.w),a5
.stay
	ret

; movef (rx),ry
dob_movef
	bne.s	.long
	move.b	(a5)+,d0
	move.b	(a5)+,d1
	move.l	(a6,d0.w),a4
	move.w	(a4),2(a6,d1.w)
	ret
.long
	move.b	(a5)+,d0
	move.b	(a5)+,d1
	move.l	(a6,d0.w),a4
	move.l	(a4),(a6,d1.w)
	ret

; movet rx,(ry)
dob_movet
	bne.s	.long
	move.b	(a5)+,d0
	move.b	(a5)+,d1
	move.l	(a6,d1.w),a4
	move.w	2(a6,d0.w),(a4)
	ret
.long
	move.b	(a5)+,d0
	move.b	(a5)+,d1
	move.l	(a6,d1.w),a4
	move.l	(a6,d0.w),(a4)
	ret

dob_jsr
	move.l	(a6,d1.w),-(sp)
	ret

;; end of ob routines


;; now the i_	macros



i_size	macro
	ifc	'\0','w'
	dc.b	0
	elseif
	dc.b	1
	endc
	endm


i_lea	macro	value,reg
	dc.b	ob_load
	dc.b	\2*4
	dc.l	\1
	endm

i_sub	macro	reg,reg
	dc.b	ob_sub
	i_size.\0
	dc.b	\1*4,\2*4
	endm

i_move	macro	reg,reg
	dc.b	ob_move
	i_size.\0
	dc.b	\1*4,\2*4
	endm

i_bne	macro	reg,reg
	dc.b	ob_bne
	i_size.\0
	dc.b	\1*4,\2*4
	endm

i_movef	macro	reg,reg
	dc.b	ob_movef
	i_size.\0
	dc.b	\1*4,\2*4
	endm

i_movet	macro	reg,reg
	dc.b	ob_movet
	i_size.\0
	dc.b	\1*4,\2*4
	endm

i_end	macro
	dc.b	ob_end,0
	endm

i_jsr	macro	rx
	dc.b	ob_jsr
	dc.b	\1*4
	endm

; that's all the basic instructions, now for the composite ones
; 	

; 14 is used as temp in composites using no other composites
; 13 is temp on composites using level 1 composites
; 12 ..... etc

i_neg	macro	rx			; uses 14
	i_lea		0,14		; r14 = 0
	i_sub.\0	\1,14		; r14 = 0 - rx
	i_move.\0	14,\1		; rx = - rx
	endm
	

i_add	macro	rx,ry			; uses 13,14
	i_move.\0	\1,13		; r13 = rx
	i_neg.\0	13		; r13 = -rx
	i_sub.\0	13,\2		; ry = ry--rx = ry+rx
	endm

i_bra	macro	addr			; use 13,14
	i_lea	\1,14
	i_lea	$0401060103,13		; jest a confusing NE number
	i_bne	13,14
	endm	

i_suba	macro	value,reg		; subtract absolute, uses 14
	i_lea	\1,14			
	i_sub	14,\2
	endm

i_adda	macro	value,reg		; add absolute, uses 12,13,14
	i_lea		-(\1),12
	i_sub		12,\2
	endm

; an obscure LEA instruction, position dependent
i_leo1	macro	value,reg		; uses 11,12,13,14
.x\@
	i_lea		.x\@-\1,11
	i_lea		.x\@,12
	i_neg.l		12
	i_add.l		12,11
	i_neg.l		11
	i_move.l	11,\2
	endm		

; another one that varies depending on where it was assembled
i_leo2	macro	value,reg
	i_lea	(\1+.x\@),\2
	i_suba	.x\@,\2
.x\@
	endm

		

ob_program
; example checksum program, works ok. about 2k per second - ok
	i_leo1	main,0
	i_lea	1,1		; constants for speed
	i_lea	2,2
	i_lea	100,3		; counter
	i_lea	0,6		; 6 = checksum
	i_lea	.next,5		; loop
	
.next
	i_movef.w	0,4
	i_add.w		4,6
	i_add.l		2,0
	i_sub.l		1,3
	i_bne		3,5
; here r6 = checksum
	i_end		

	even
ob_struct
	ds.b	ob_len
	even
